﻿using System;
using System.Collections.Generic;
using UnityEngine;
using Newtonsoft.Json;
using System.Linq;
using REST_JSON_API;
using System.Linq.Expressions;

public class FunctionCallFact : FactWrappedCRTP<FunctionCallFact>
{
    public string func_id;

    public string arg_func_id;

    public (float t_0, float t_n) Domain;


    [JsonIgnore]
    public FunctionFact Function
        => (FunctionFact)FactRecorder.AllFacts[func_id];

    [JsonIgnore]
    public FunctionFact Function_args
        => (FunctionFact)FactRecorder.AllFacts[arg_func_id];


    public FunctionCallFact() : base() { }

    public FunctionCallFact(string func_id, string arg_func_id, (float t_0, float t_n) Domain, FactRecorder organizer) : base(organizer)
    {
        this.func_id = func_id;
        this.arg_func_id = arg_func_id;
        this.Domain = Domain;

        ServerDefinition = new OMS(func_id.ToString() + arg_func_id.ToString() + Domain.ToString());
    }

    public object[] Call(float t)
    {
        if (t < Domain.t_0 || t > Domain.t_n)
            return null;

        object[] path = Function_args.Function(new object[] { t });
        return Function.Function(path);
    }

    public override bool HasDependentFacts
        => true;

    protected override string[] GetDependentFactIds()
        => new[] { func_id, arg_func_id };

    protected override bool EquivalentWrapped(FunctionCallFact f1, FunctionCallFact f2)
    {
        if (Mathf.Approximately(f1.Domain.t_0, f2.Domain.t_0)
            && Mathf.Approximately(f1.Domain.t_n, f2.Domain.t_n)
            && DependentFactsEquivalent(f1, f2)
        )
            return true;

        if (f1.Function.Signature[^1] != f2.Function.Signature[^1])
            return false;

        int samplenumber = 3;
        float stepsizef1 = (f1.Domain.t_n - f1.Domain.t_0) / (samplenumber - 1);
        float stepsizef2 = (f2.Domain.t_n - f2.Domain.t_0) / (samplenumber - 1);
        for ((int i, float f1step, float f2step) = (0, f1.Domain.t_0, f2.Domain.t_0);
            i < samplenumber;
            i++, f1step += stepsizef1, f2step += stepsizef2)
        {
            object[] c1 = f1.Call(f1step);
            object[] c2 = f2.Call(f2step);
            if (c1 == null || c2 == null
            || !c1.SequenceEqual(c2, new ApproximationComparer()))
                return false;
        }
        return true;
    }

    public override MMTFact MakeMMTDeclaration()
    {
        MMTGeneralFact mmt_arg = (MMTGeneralFact)Function_args.MakeMMTDeclaration();
        MMTGeneralFact mmt_in = (MMTGeneralFact)Function.MakeMMTDeclaration();

        SOMDoc type = new OMA(
            new OMS(MMTConstants.Product),
            new[]{
                new OMA(
                    new OMS(MMTConstants.Product),
                    new []{
                        new OMS(MMTConstants.RealLit),
                        new OMS(MMTConstants.RealLit)
                }),
                mmt_arg.type,
                mmt_in.type,
            });

        SOMDoc defines = new OMA(
            new OMS(MMTConstants.Tuple),
            new[] {
                new OMA(
                    new OMS(MMTConstants.Tuple),
                    new[]{
                        new OMLIT<float>(Domain.t_0),
                        new OMLIT<float>(Domain.t_n)
                }),
                Function_args.ServerDefinition,
                Function.ServerDefinition,
        });

        return new MMTGeneralFact(Label, type, defines);
    }

    protected override void RecalculateTransform()
    {
        Position = Function.Position;
        Rotation = Function.Rotation;
        LocalScale = Function.LocalScale;
    }

    protected override Fact _ReInitializeMe(Dictionary<string, string> old_to_new, FactRecorder organizer)
        => new FunctionCallFact(old_to_new[this.func_id], old_to_new[this.arg_func_id], this.Domain, organizer);
}

public class FunctionFact : FactWrappedCRTP<FunctionFact>
{
    public SOMDoc Function_SOMDoc;

    [JsonIgnore]
    public Type[] Signature;

    [JsonIgnore]
    public Func<object[], object[]> Function;

    [JsonIgnore]
    public Expression LambdaExpression;

    /// <summary> \copydoc Fact.Fact </summary>
    public FunctionFact() : base() { }

    /// <summary>
    /// Standard Constructor:
    /// Initiates members and creates MMT %Fact Server-Side
    /// </summary>
    /// <param name="Function_SOMDoc">sets <see cref="Function_SOMDoc"/> and contains the Abstract Syntax Tree</param>
    /// <param name="organizer">sets <see cref="Fact._Facts"/></param>
    public FunctionFact(SOMDoc Function_SOMDoc, FactRecorder organizer) : base(organizer)
    {
        this.Function_SOMDoc = Function_SOMDoc;
        this.Function = this.Function_SOMDoc.PartialInvokeCastingLambdaExpression(out LambdaExpression, out Signature);
        SendToMMT();
    }

    /// <summary>
    /// Bypasses initialization of new MMT %Fact by using existend URI, _which is not checked for existence_.
    /// </summary>
    /// <param name="function_expression">sets <see cref="Function_expression"/> and contains the Abstract Syntax Tree</param>
    /// <param name="uri">MMT URI</param>
    /// <param name="organizer">sets <see cref="Fact._Facts"/></param>
    public FunctionFact(SOMDoc Function_SOMDoc, SOMDoc _ServerDefinition, FactRecorder organizer) : base(organizer)
    {
        this.Function_SOMDoc = Function_SOMDoc;
        this.Function = Function_SOMDoc.PartialInvokeCastingLambdaExpression(out LambdaExpression, out Signature);

        this.ServerDefinition = _ServerDefinition;
        _ = this.Label;
    }

    protected override void RecalculateTransform() { }

    /// \copydoc Fact.parseFact(ScrollFact)
    public new static List<Fact> parseFact(MMTFact fact)
    {// TODO Correctness

        if (((MMTGeneralFact)fact).defines is not FUN fun
            && (((MMTGeneralFact)fact).defines is not FallbackWrapper SFW
            || (fun = SFW.SFunction as FUN) == null)
        )
            return new();

        ParsingDictionary.parseTermsToId.TryAdd(fun.ToString(), fact.@ref.uri);

        return new() { new FunctionFact(fun, fact.@ref, StageStatic.stage.factState) };
    }

    /// \copydoc Fact.hasDependentFacts
    public override bool HasDependentFacts
        => false;

    /// \copydoc Fact.getDependentFactIds
    protected override string[] GetDependentFactIds()
        => new string[] { };

    /// \copydoc Fact.GetHashCode
    public override int GetHashCode()
        => Function.GetHashCode();

    /// \copydoc Fact.EquivalentWrapped
    protected override bool EquivalentWrapped(FunctionFact f1, FunctionFact f2)
        => f1.Function_SOMDoc.Equivalent(f2.Function_SOMDoc);

    protected override Fact _ReInitializeMe(Dictionary<string, string> old_to_new, FactRecorder organizer)
        => new FunctionFact(this.Function_SOMDoc.MapURIs(old_to_new), organizer);

    public override MMTFact MakeMMTDeclaration()
        => new MMTGeneralFact(
            Label,
            SOMDoc.SOMDocType(FuncExtensions.CreateFuncType(Signature)),
            Function_SOMDoc);
}